"""
Google's clitable.py is inherently integrated to Linux:

This is a workaround for that (basically include modified clitable code without anything
that is Linux-specific).

_clitable.py is identical to Google's as of 2017-12-17
_texttable.py is identical to Google's as of 2017-12-17
_terminal.py is a highly stripped down version of Google's such that clitable.py works

https://github.com/google/textfsm/blob/master/clitable.py

A module to represent and manipulate tabular text data.

A table of rows, indexed on row number. Each row is a ordered dictionary of row
elements that maintains knowledge of the parent table and column headings.

Tables can be created from CSV input and in-turn supports a number of display
formats such as CSV and variable sized and justified rows.
"""

# Some of this code is from Google with the following license:
#
# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import copy
from functools import cmp_to_key
import textwrap

# pylint: disable=redefined-builtin
from six.moves import range
from netmiko._textfsm import _terminal as terminal


class Error(Exception):
    """Base class for errors."""


class TableError(Error):
    """Error in TextTable."""


class Row(dict):
    """Represents a table row. We implement this as an ordered dictionary.

  The order is the chronological order of data insertion. Methods are supplied
  to make it behave like a regular dict() and list().

  Attributes:
    row: int, the row number in the container table. 0 is the header row.
    table: A TextTable(), the associated container table.
  """

    def __init__(self, *args, **kwargs):
        super(Row, self).__init__(*args, **kwargs)
        self._keys = list()
        self._values = list()
        self.row = None
        self.table = None
        self._color = None
        self._index = {}

    def _BuildIndex(self):
        """Recreate the key index."""
        self._index = {}
        for i, k in enumerate(self._keys):
            self._index[k] = i

    def __getitem__(self, column):
        """Support for [] notation.

    Args:
      column: Tuple of column names, or a (str) column name, or positional
        column number, 0-indexed.

    Returns:
      A list or string with column value(s).

    Raises:
      IndexError: The given column(s) were not found.
    """
        if isinstance(column, (list, tuple)):
            ret = []
            for col in column:
                ret.append(self[col])
            return ret

        try:
            return self._values[self._index[column]]
        except (KeyError, TypeError, ValueError):
            pass

        # Perhaps we have a range like '1', ':-1' or '1:'.
        try:
            return self._values[column]
        except (IndexError, TypeError):
            pass

        raise IndexError('No such column "%s" in row.' % column)

    def __contains__(self, value):
        return value in self._values

    def __setitem__(self, column, value):
        for i in range(len(self)):
            if self._keys[i] == column:
                self._values[i] = value
                return
        # No column found, add a new one.
        self._keys.append(column)
        self._values.append(value)
        self._BuildIndex()

    def __iter__(self):
        return iter(self._values)

    def __len__(self):
        return len(self._keys)

    def __str__(self):
        ret = ""
        for v in self._values:
            ret += "%12s  " % v
        ret += "\n"
        return ret

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, str(self))

    def get(self, column, default_value=None):
        """Get an item from the Row by column name.

    Args:
      column: Tuple of column names, or a (str) column name, or positional
        column number, 0-indexed.
      default_value: The value to use if the key is not found.

    Returns:
      A list or string with column value(s) or default_value if not found.
    """
        if isinstance(column, (list, tuple)):
            ret = []
            for col in column:
                ret.append(self.get(col, default_value))
            return ret
        # Perhaps we have a range like '1', ':-1' or '1:'.
        try:
            return self._values[column]
        except (IndexError, TypeError):
            pass
        try:
            return self[column]
        except IndexError:
            return default_value

    def index(self, column):  # pylint: disable=C6409
        """Fetches the column number (0 indexed).

    Args:
      column: A string, column to fetch the index of.

    Returns:
      An int, the row index number.

    Raises:
      ValueError: The specified column was not found.
    """
        for i, key in enumerate(self._keys):
            if key == column:
                return i
        raise ValueError('Column "%s" not found.' % column)

    def iterkeys(self):
        return iter(self._keys)

    def items(self):
        # TODO(harro): self.get(k) should work here but didn't ?
        return [(k, self.__getitem__(k)) for k in self._keys]

    def _GetValues(self):
        """Return the row's values."""
        return self._values

    def _GetHeader(self):
        """Return the row's header."""
        return self._keys

    def _SetHeader(self, values):
        """Set the row's header from a list."""
        if self._values and len(values) != len(self._values):
            raise ValueError("Header values not equal to existing data width.")
        if not self._values:
            for _ in range(len(values)):
                self._values.append(None)
        self._keys = list(values)
        self._BuildIndex()

    def _SetColour(self, value_list):
        """Sets row's colour attributes to a list of values in terminal.SGR."""
        if value_list is None:
            self._color = None
            return
        colors = []
        for color in value_list:
            if color in terminal.SGR:
                colors.append(color)
            elif color in terminal.FG_COLOR_WORDS:
                colors += terminal.FG_COLOR_WORDS[color]
            elif color in terminal.BG_COLOR_WORDS:
                colors += terminal.BG_COLOR_WORDS[color]
            else:
                raise ValueError("Invalid colour specification.")
        self._color = list(set(colors))

    def _GetColour(self):
        if self._color is None:
            return None
        return list(self._color)

    def _SetValues(self, values):
        """Set values from supplied dictionary or list.

    Args:
      values: A Row, dict indexed by column name, or list.

    Raises:
      TypeError: Argument is not a list or dict, or list is not equal row
      length or dictionary keys don't match.
    """

        def _ToStr(value):
            """Convert individul list entries to string."""
            if isinstance(value, (list, tuple)):
                result = []
                for val in value:
                    result.append(str(val))
                return result
            else:
                return str(value)

        # Row with identical header can be copied directly.
        if isinstance(values, Row):
            if self._keys != values.header:
                raise TypeError("Attempt to append row with mismatched header.")
            self._values = copy.deepcopy(values.values)

        elif isinstance(values, dict):
            for key in self._keys:
                if key not in values:
                    raise TypeError("Dictionary key mismatch with row.")
            for key in self._keys:
                self[key] = _ToStr(values[key])

        elif isinstance(values, list) or isinstance(values, tuple):
            if len(values) != len(self._values):
                raise TypeError("Supplied list length != row length")
            for (index, value) in enumerate(values):
                self._values[index] = _ToStr(value)

        else:
            raise TypeError(
                "Supplied argument must be Row, dict or list, not %s", type(values)
            )

    def Insert(self, key, value, row_index):
        """Inserts new values at a specified offset.

    Args:
      key: string for header value.
      value: string for a data value.
      row_index: Offset into row for data.

    Raises:
      IndexError: If the offset is out of bands.
    """
        if row_index < 0:
            row_index += len(self)

        if not 0 <= row_index < len(self):
            raise IndexError('Index "%s" is out of bounds.' % row_index)

        new_row = Row()
        for idx in self.header:
            if self.index(idx) == row_index:
                new_row[key] = value
            new_row[idx] = self[idx]
        self._keys = new_row.header
        self._values = new_row.values
        del new_row
        self._BuildIndex()

    color = property(_GetColour, _SetColour, doc="Colour spec of this row")
    header = property(_GetHeader, _SetHeader, doc="List of row's headers.")
    values = property(_GetValues, _SetValues, doc="List of row's values.")


class TextTable(object):
    """Class that provides data methods on a tabular format.

  Data is stored as a list of Row() objects. The first row is always present as
  the header row.

  Attributes:
    row_class: class, A class to use for the Row object.
    separator: str, field separator when printing table.
  """

    def __init__(self, row_class=Row):
        """Initialises a new table.

    Args:
      row_class: A class to use as the row object. This should be a
        subclass of this module's Row() class.
    """
        self.row_class = row_class
        self.separator = ", "
        self.Reset()

    def Reset(self):
        self._row_index = 1
        self._table = [[]]
        self._iterator = 0  # While loop row index

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, str(self))

    def __str__(self):
        """Displays table with pretty formatting."""
        return self.table

    def __incr__(self, incr=1):
        self._SetRowIndex(self._row_index + incr)

    def __contains__(self, name):
        """Whether the given column header name exists."""
        return name in self.header

    def __getitem__(self, row):
        """Fetches the given row number."""
        return self._table[row]

    def __iter__(self):
        """Iterator that excludes the header row."""
        return self.next()

    def next(self):
        # Maintain a counter so a row can know what index it is.
        # Save the old value to support nested interations.
        old_iter = self._iterator
        try:
            for r in self._table[1:]:
                self._iterator = r.row
                yield r
        finally:
            # Recover the original index after loop termination or exit with break.
            self._iterator = old_iter

    def __add__(self, other):
        """Merges two with identical columns."""

        new_table = copy.copy(self)
        for row in other:
            new_table.Append(row)

        return new_table

    def __copy__(self):
        """Copy table instance."""

        new_table = self.__class__()
        # pylint: disable=protected-access
        new_table._table = [self.header]
        for row in self[1:]:
            new_table.Append(row)
        return new_table

    def Filter(self, function=None):
        """Construct Textable from the rows of which the function returns true.


    Args:
      function: A function applied to each row which returns a bool. If
                function is None, all rows with empty column values are
                removed.
    Returns:
      A new TextTable()

    Raises:
      TableError: When an invalid row entry is Append()'d
    """
        flat = (
            lambda x: x if isinstance(x, str) else "".join([flat(y) for y in x])
        )  # noqa
        if function is None:
            function = lambda row: bool(flat(row.values))  # noqa

        new_table = self.__class__()
        # pylint: disable=protected-access
        new_table._table = [self.header]
        for row in self:
            if function(row) is True:
                new_table.Append(row)
        return new_table

    def Map(self, function):
        """Applies the function to every row in the table.

    Args:
      function: A function applied to each row.

    Returns:
      A new TextTable()

    Raises:
      TableError: When transform is not invalid row entry. The transform
                  must be compatible with Append().
    """
        new_table = self.__class__()
        # pylint: disable=protected-access
        new_table._table = [self.header]
        for row in self:
            filtered_row = function(row)
            if filtered_row:
                new_table.Append(filtered_row)
        return new_table

    # pylint: disable=C6409
    # pylint: disable=W0622
    def sort(self, cmp=None, key=None, reverse=False):
        """Sorts rows in the texttable.

    Args:
      cmp: func, non default sort algorithm to use.
      key: func, applied to each element before sorting.
      reverse: bool, reverse order of sort.
    """

        def _DefaultKey(value):
            """Default key func is to create a list of all fields."""
            result = []
            for key in self.header:
                # Try sorting as numerical value if possible.
                try:
                    result.append(float(value[key]))
                except ValueError:
                    result.append(value[key])
            return result

        key = key or _DefaultKey
        # Exclude header by copying table.
        new_table = self._table[1:]

        if cmp is not None:
            key = cmp_to_key(cmp)

        new_table.sort(key=key, reverse=reverse)

        # Regenerate the table with original header
        self._table = [self.header]
        self._table.extend(new_table)
        # Re-write the 'row' attribute of each row
        for index, row in enumerate(self._table):
            row.row = index

    # pylint: enable=W0622

    def extend(self, table, keys=None):
        """Extends all rows in the texttable.

    The rows are extended with the new columns from the table.

    Args:
      table: A texttable, the table to extend this table by.
      keys: A set, the set of columns to use as the key. If None, the
        row index is used.

    Raises:
      IndexError: If key is not a valid column name.
    """
        if keys:
            for k in keys:
                if k not in self._Header():
                    raise IndexError("Unknown key: '%s'", k)

        extend_with = []
        for column in table.header:
            if column not in self.header:
                extend_with.append(column)

        if not extend_with:
            return

        for column in extend_with:
            self.AddColumn(column)

        if not keys:
            for row1, row2 in zip(self, table):
                for column in extend_with:
                    row1[column] = row2[column]
            return

        for row1 in self:
            for row2 in table:
                for k in keys:
                    if row1[k] != row2[k]:
                        break
                else:
                    for column in extend_with:
                        row1[column] = row2[column]
                    break

    # pylint: enable=C6409
    def Remove(self, row):
        """Removes a row from the table.

    Args:
      row: int, the row number to delete. Must be >= 1, as the header
        cannot be removed.

    Raises:
      TableError: Attempt to remove nonexistent or header row.
    """
        if row == 0 or row > self.size:
            raise TableError("Attempt to remove header row")
        new_table = []
        # pylint: disable=E1103
        for t_row in self._table:
            if t_row.row != row:
                new_table.append(t_row)
                if t_row.row > row:
                    t_row.row -= 1
        self._table = new_table

    def _Header(self):
        """Returns the header row."""
        return self._table[0]

    def _GetRow(self, columns=None):
        """Returns the current row as a tuple."""

        row = self._table[self._row_index]
        if columns:
            result = []
            for col in columns:
                if col not in self.header:
                    raise TableError("Column header %s not known in table." % col)
                result.append(row[self.header.index(col)])
            row = result
        return row

    def _SetRow(self, new_values, row=0):
        """Sets the current row to new list.

    Args:
      new_values: List|dict of new values to insert into row.
      row: int, Row to insert values into.

    Raises:
      TableError: If number of new values is not equal to row size.
    """

        if not row:
            row = self._row_index

        if row > self.size:
            raise TableError("Entry %s beyond table size %s." % (row, self.size))

        self._table[row].values = new_values

    def _SetHeader(self, new_values):
        """Sets header of table to the given tuple.

    Args:
      new_values: Tuple of new header values.
    """
        row = self.row_class()
        row.row = 0
        for v in new_values:
            row[v] = v
        self._table[0] = row

    def _SetRowIndex(self, row):
        if not row or row > self.size:
            raise TableError("Entry %s beyond table size %s." % (row, self.size))
        self._row_index = row

    def _GetRowIndex(self):
        return self._row_index

    def _GetSize(self):
        """Returns number of rows in table."""

        if not self._table:
            return 0
        return len(self._table) - 1

    def _GetTable(self):
        """Returns table, with column headers and separators.

    Returns:
      The whole table including headers as a string. Each row is
      joined by a newline and each entry by self.separator.
    """
        result = []
        # Avoid the global lookup cost on each iteration.
        lstr = str
        for row in self._table:
            result.append("%s\n" % self.separator.join(lstr(v) for v in row))

        return "".join(result)

    def _SetTable(self, table):
        """Sets table, with column headers and separators."""
        if not isinstance(table, TextTable):
            raise TypeError("Not an instance of TextTable.")
        self.Reset()
        self._table = copy.deepcopy(table._table)  # pylint: disable=W0212
        # Point parent table of each row back ourselves.
        for row in self:
            row.table = self

    def _SmallestColSize(self, text):
        """Finds the largest indivisible word of a string.

    ...and thus the smallest possible column width that can contain that
    word unsplit over rows.

    Args:
      text: A string of text potentially consisting of words.

    Returns:
      Integer size of the largest single word in the text.
    """
        if not text:
            return 0
        stripped = terminal.StripAnsiText(text)
        return max(len(word) for word in stripped.split())

    def _TextJustify(self, text, col_size):
        """Formats text within column with white space padding.

    A single space is prefixed, and a number of spaces are added as a
    suffix such that the length of the resultant string equals the col_size.

    If the length of the text exceeds the column width available then it
    is split into words and returned as a list of string, each string
    contains one or more words padded to the column size.

    Args:
      text: String of text to format.
      col_size: integer size of column to pad out the text to.

    Returns:
      List of strings col_size in length.

    Raises:
      TableError: If col_size is too small to fit the words in the text.
    """
        result = []
        if "\n" in text:
            for paragraph in text.split("\n"):
                result.extend(self._TextJustify(paragraph, col_size))
            return result

        wrapper = textwrap.TextWrapper(
            width=col_size - 2, break_long_words=False, expand_tabs=False
        )
        try:
            text_list = wrapper.wrap(text)
        except ValueError:
            raise TableError("Field too small (minimum width: 3)")

        if not text_list:
            return [" " * col_size]

        for current_line in text_list:
            stripped_len = len(terminal.StripAnsiText(current_line))
            ansi_color_adds = len(current_line) - stripped_len
            # +2 for white space on either side.
            if stripped_len + 2 > col_size:
                raise TableError("String contains words that do not fit in column.")

            result.append(" %-*s" % (col_size - 1 + ansi_color_adds, current_line))

        return result

    def FormattedTable(
        self,
        width=80,
        force_display=False,
        ml_delimiter=True,
        color=True,
        display_header=True,
        columns=None,
    ):
        """Returns whole table, with whitespace padding and row delimiters.

    Args:
      width: An int, the max width we want the table to fit in.
      force_display: A bool, if set to True will display table when the table
          can't be made to fit to the width.
      ml_delimiter: A bool, if set to False will not display the multi-line
          delimiter.
      color: A bool. If true, display any colours in row.colour.
      display_header: A bool. If true, display header.
      columns: A list of str, show only columns with these names.

    Returns:
      A string.  The tabled output.

    Raises:
      TableError: Width too narrow to display table.
    """

        def _FilteredCols():
            """Returns list of column names to display."""
            if not columns:
                return self._Header().values
            return [col for col in self._Header().values if col in columns]

        # Largest is the biggest data entry in a column.
        largest = {}
        # Smallest is the same as above but with linewrap i.e. largest unbroken
        # word in the data stream.
        smallest = {}
        # largest == smallest for a column with a single word of data.
        # Initialise largest and smallest for all columns.
        for key in _FilteredCols():
            largest[key] = 0
            smallest[key] = 0

        # Find the largest and smallest values.
        # Include Title line in equation.
        # pylint: disable=E1103
        for row in self._table:
            for key, value in row.items():
                if key not in _FilteredCols():
                    continue
                # Convert lists into a string.
                if isinstance(value, list):
                    value = ", ".join(value)
                value = terminal.StripAnsiText(value)
                largest[key] = max(len(value), largest[key])
                smallest[key] = max(self._SmallestColSize(value), smallest[key])
        # pylint: enable=E1103

        min_total_width = 0
        multi_word = []
        # Bump up the size of each column to include minimum pad.
        # Find all columns that can be wrapped (multi-line).
        # And the minimum width needed to display all columns (even if wrapped).
        for key in _FilteredCols():
            # Each column is bracketed by a space on both sides.
            # So increase size required accordingly.
            largest[key] += 2
            smallest[key] += 2
            min_total_width += smallest[key]
            # If column contains data that 'could' be split over multiple lines.
            if largest[key] != smallest[key]:
                multi_word.append(key)

        # Check if we have enough space to display the table.
        if min_total_width > width and not force_display:
            raise TableError("Width too narrow to display table.")

        # We have some columns that may need wrapping over several lines.
        if multi_word:
            # Find how much space is left over for the wrapped columns to use.
            # Also find how much space we would need if they were not wrapped.
            # These are 'spare_width' and 'desired_width' respectively.
            desired_width = 0
            spare_width = width - min_total_width
            for key in multi_word:
                spare_width += smallest[key]
                desired_width += largest[key]

            # Scale up the space we give each wrapped column.
            # Proportional to its size relative to 'desired_width' for all columns.
            # Rinse and repeat if we changed the wrap list in this iteration.
            # Once done we will have a list of columns that definitely need wrapping.
            done = False
            while not done:
                done = True
                for key in multi_word:
                    # If we scale past the desired width for this particular column,
                    # then give it its desired width and remove it from the wrapped list.
                    if largest[key] <= round(
                        (largest[key] / float(desired_width)) * spare_width
                    ):
                        smallest[key] = largest[key]
                        multi_word.remove(key)
                        spare_width -= smallest[key]
                        desired_width -= largest[key]
                        done = False
                    # If we scale below the minimum width for this particular column,
                    # then leave it at its minimum and remove it from the wrapped list.
                    elif smallest[key] >= round(
                        (largest[key] / float(desired_width)) * spare_width
                    ):
                        multi_word.remove(key)
                        spare_width -= smallest[key]
                        desired_width -= largest[key]
                        done = False

            # Repeat the scaling algorithm with the final wrap list.
            # This time we assign the extra column space by increasing 'smallest'.
            for key in multi_word:
                smallest[key] = int(
                    round((largest[key] / float(desired_width)) * spare_width)
                )

        total_width = 0
        row_count = 0
        result_dict = {}
        # Format the header lines and add to result_dict.
        # Find what the total width will be and use this for the ruled lines.
        # Find how many rows are needed for the most wrapped line (row_count).
        for key in _FilteredCols():
            result_dict[key] = self._TextJustify(key, smallest[key])
            if len(result_dict[key]) > row_count:
                row_count = len(result_dict[key])
            total_width += smallest[key]

        # Store header in header_list, working down the wrapped rows.
        header_list = []
        for row_idx in range(row_count):
            for key in _FilteredCols():
                try:
                    header_list.append(result_dict[key][row_idx])
                except IndexError:
                    # If no value than use whitespace of equal size.
                    header_list.append(" " * smallest[key])
            header_list.append("\n")

        # Format and store the body lines
        result_dict = {}
        body_list = []
        # We separate multi line rows with a single line delimiter.
        prev_muli_line = False
        # Unless it is the first line in which there is already the header line.
        first_line = True
        for row in self:
            row_count = 0
            for key, value in row.items():
                if key not in _FilteredCols():
                    continue
                # Convert field contents to a string.
                if isinstance(value, list):
                    value = ", ".join(value)
                # Store results in result_dict and take note of wrapped line count.
                result_dict[key] = self._TextJustify(value, smallest[key])
                if len(result_dict[key]) > row_count:
                    row_count = len(result_dict[key])

            if row_count > 1:
                prev_muli_line = True
            # If current or prior line was multi-line then include delimiter.
            if not first_line and prev_muli_line and ml_delimiter:
                body_list.append("-" * total_width + "\n")
                if row_count == 1:
                    # Our current line was not wrapped, so clear flag.
                    prev_muli_line = False

            row_list = []
            for row_idx in range(row_count):
                for key in _FilteredCols():
                    try:
                        row_list.append(result_dict[key][row_idx])
                    except IndexError:
                        # If no value than use whitespace of equal size.
                        row_list.append(" " * smallest[key])
                row_list.append("\n")

            if color and row.color is not None:
                # Don't care about colors
                body_list.append("".join(row_list))
                # body_list.append(
                #    terminal.AnsiText(''.join(row_list)[:-1],
                #                      command_list=row.color))
                # body_list.append('\n')
            else:
                body_list.append("".join(row_list))

            first_line = False

        header = "".join(header_list) + "=" * total_width
        if color and self._Header().color is not None:
            pass
            # header = terminal.AnsiText(header, command_list=self._Header().color)
        # Add double line delimiter between header and main body.
        if display_header:
            return "%s\n%s" % (header, "".join(body_list))
        return "%s" % "".join(body_list)

    def LabelValueTable(self, label_list=None):
        """Returns whole table as rows of name/value pairs.

    One (or more) column entries are used for the row prefix label.
    The remaining columns are each displayed as a row entry with the
    prefix labels appended.

    Use the first column as the label if label_list is None.

    Args:
      label_list: A list of prefix labels to use.

    Returns:
      Label/Value formatted table.

    Raises:
      TableError: If specified label is not a column header of the table.
    """
        label_list = label_list or self._Header()[0]
        # Ensure all labels are valid.
        for label in label_list:
            if label not in self._Header():
                raise TableError("Invalid label prefix: %s." % label)

        sorted_list = []
        for header in self._Header():
            if header in label_list:
                sorted_list.append(header)

        label_str = "# LABEL %s\n" % ".".join(sorted_list)

        body = []
        for row in self:
            # Some of the row values are pulled into the label, stored in label_prefix.
            label_prefix = []
            value_list = []
            for key, value in row.items():
                if key in sorted_list:
                    # Set prefix.
                    label_prefix.append(value)
                else:
                    value_list.append("%s %s" % (key, value))

            body.append(
                "".join(["%s.%s\n" % (".".join(label_prefix), v) for v in value_list])
            )

        return "%s%s" % (label_str, "".join(body))

    table = property(_GetTable, _SetTable, doc="Whole table")
    row = property(_GetRow, _SetRow, doc="Current row")
    header = property(_Header, _SetHeader, doc="List of header entries.")
    row_index = property(_GetRowIndex, _SetRowIndex, doc="Current row.")
    size = property(_GetSize, doc="Number of rows in table.")

    def RowWith(self, column, value):
        """Retrieves the first non header row with the column of the given value.

    Args:
      column: str, the name of the column to check.
      value: str, The value of the column to check.

    Returns:
      A Row() of the first row found, None otherwise.

    Raises:
      IndexError: The specified column does not exist.
    """
        for row in self._table[1:]:
            if row[column] == value:
                return row
        return None

    def AddColumn(self, column, default="", col_index=-1):
        """Appends a new column to the table.

    Args:
      column: A string, name of the column to add.
      default: Default value for entries. Defaults to ''.
      col_index: Integer index for where to insert new column.

    Raises:
      TableError: Column name already exists.

    """
        if column in self.table:
            raise TableError("Column %r already in table." % column)
        if col_index == -1:
            self._table[0][column] = column
            for i in range(1, len(self._table)):
                self._table[i][column] = default
        else:
            self._table[0].Insert(column, column, col_index)
            for i in range(1, len(self._table)):
                self._table[i].Insert(column, default, col_index)

    def Append(self, new_values):
        """Adds a new row (list) to the table.

    Args:
      new_values: Tuple, dict, or Row() of new values to append as a row.

    Raises:
      TableError: Supplied tuple not equal to table width.
    """
        newrow = self.NewRow()
        newrow.values = new_values
        self._table.append(newrow)

    def NewRow(self, value=""):
        """Fetches a new, empty row, with headers populated.

    Args:
      value: Initial value to set each row entry to.

    Returns:
      A Row() object.
    """
        newrow = self.row_class()
        newrow.row = self.size + 1
        newrow.table = self
        headers = self._Header()
        for header in headers:
            newrow[header] = value
        return newrow

    def CsvToTable(self, buf, header=True, separator=","):
        """Parses buffer into tabular format.

    Strips off comments (preceded by '#').
    Optionally parses and indexes by first line (header).

    Args:
      buf: String file buffer containing CSV data.
      header: Is the first line of buffer a header.
      separator: String that CSV is separated by.

    Returns:
      int, the size of the table created.

    Raises:
      TableError: A parsing error occurred.
    """
        self.Reset()

        header_row = self.row_class()
        if header:
            line = buf.readline()
            header_str = ""
            while not header_str:
                # Remove comments.
                header_str = line.split("#")[0].strip()
                if not header_str:
                    line = buf.readline()

            header_list = header_str.split(separator)
            header_length = len(header_list)

            for entry in header_list:
                entry = entry.strip()
                if entry in header_row:
                    raise TableError("Duplicate header entry %r." % entry)

                header_row[entry] = entry
            header_row.row = 0
            self._table[0] = header_row

        # xreadlines would be better but not supported by StringIO for testing.
        for line in buf:
            # Support commented lines, provide '#' is first character of line.
            if line.startswith("#"):
                continue

            lst = line.split(separator)
            lst = [l.strip() for l in lst]
            if header and len(lst) != header_length:
                # Silently drop illegal line entries
                continue
            if not header:
                header_row = self.row_class()
                header_length = len(lst)
                header_row.values = dict(
                    zip(range(header_length), range(header_length))
                )
                self._table[0] = header_row
                header = True
                continue

            new_row = self.NewRow()
            new_row.values = lst
            header_row.row = self.size + 1
            self._table.append(new_row)

        return self.size

    def index(self, name=None):  # pylint: disable=C6409
        """Returns index number of supplied column name.

    Args:
      name: string of column name.

    Raises:
      TableError: If name not found.

    Returns:
      Index of the specified header entry.
    """
        try:
            return self.header.index(name)
        except ValueError:
            raise TableError("Unknown index name %s." % name)



import os
import time
import pkg_resources
import shutil

from preventivemaintenance.script.capture import function_capture

capture_path = pkg_resources.resource_filename('preventivemaintenance', 'capture/')
data_path = pkg_resources.resource_filename('preventivemaintenance', 'data/')
result_path = pkg_resources.resource_filename('preventivemaintenance', 'result/')

print(capture_path)

answer=input(
    'Press Number for Menu\n'
    '1. Capture\n'
    '2. Create Report\n'
    '3. Analyze Capture with NTC_template(Dev)\n'
    '4. Ping(Still not Work)\n'
    'Type "quit" to quit program\n'
    )
if answer == '1':
    data_dir = (data_path+'/devices_data.xlsx')
    function_capture=function_capture(data_dir,capture_path)
elif answer == '2':
    try:
        chg_dir = os.chdir(capture_path)
        #current_dir = os.path.dirname(os.path.realpath(__file__))
        current_dir=os.getcwd()
        from preventivemaintenance.script.file_identification import file_identification
        from preventivemaintenance.script.to_docx import to_docx
        files = os.listdir(current_dir)
        file_identification=file_identification(files)
        file_identification.file_identification()
        to_docx=to_docx()
        to_docx.to_docx()
        time.sleep(3)
        src_mv = (capture_path+'preventive_maintenance.docx')
        dst_mv = (result_path+'preventive_maintenance.docx')
        shutil.move(src_mv,dst_mv)

    except NameError:
        raise
elif answer == '3':
    current_dir = os.path.dirname(os.path.realpath(__file__))
    src_mv = (current_dir+'/Capture/result.xls')
    dst_mv = (current_dir+'/Result/result.xls')
    chg_dir = os.chdir(current_dir+'/Capture')
    files = os.listdir(chg_dir)
    function_analyze_capture=function_analyze_capture(files)
    time.sleep(3)
    shutil.move(src_mv,dst_mv)


# encoding: utf-8

"""
Namespace-related objects.
"""

from __future__ import absolute_import, print_function, unicode_literals


nsmap = {
    "a": ("http://schemas.openxmlformats.org/drawingml/2006/main"),
    "c": ("http://schemas.openxmlformats.org/drawingml/2006/chart"),
    "cp": ("http://schemas.openxmlformats.org/package/2006/metadata/core-properties"),
    "dc": ("http://purl.org/dc/elements/1.1/"),
    "dcmitype": ("http://purl.org/dc/dcmitype/"),
    "dcterms": ("http://purl.org/dc/terms/"),
    "dgm": ("http://schemas.openxmlformats.org/drawingml/2006/diagram"),
    "m": ("http://schemas.openxmlformats.org/officeDocument/2006/math"),
    "pic": ("http://schemas.openxmlformats.org/drawingml/2006/picture"),
    "r": ("http://schemas.openxmlformats.org/officeDocument/2006/relationships"),
    "sl": ("http://schemas.openxmlformats.org/schemaLibrary/2006/main"),
    "w": ("http://schemas.openxmlformats.org/wordprocessingml/2006/main"),
    "wp": ("http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"),
    "xml": ("http://www.w3.org/XML/1998/namespace"),
    "xsi": ("http://www.w3.org/2001/XMLSchema-instance"),
}

pfxmap = dict((value, key) for key, value in nsmap.items())


class NamespacePrefixedTag(str):
    """
    Value object that knows the semantics of an XML tag having a namespace
    prefix.
    """
    def __new__(cls, nstag, *args):
        return super(NamespacePrefixedTag, cls).__new__(cls, nstag)

    def __init__(self, nstag):
        self._pfx, self._local_part = nstag.split(':')
        self._ns_uri = nsmap[self._pfx]

    @property
    def clark_name(self):
        return '{%s}%s' % (self._ns_uri, self._local_part)

    @classmethod
    def from_clark_name(cls, clark_name):
        nsuri, local_name = clark_name[1:].split('}')
        nstag = '%s:%s' % (pfxmap[nsuri], local_name)
        return cls(nstag)

    @property
    def local_part(self):
        """
        Return the local part of the tag as a string. E.g. 'foobar' is
        returned for tag 'f:foobar'.
        """
        return self._local_part

    @property
    def nsmap(self):
        """
        Return a dict having a single member, mapping the namespace prefix of
        this tag to it's namespace name (e.g. {'f': 'http://foo/bar'}). This
        is handy for passing to xpath calls and other uses.
        """
        return {self._pfx: self._ns_uri}

    @property
    def nspfx(self):
        """
        Return the string namespace prefix for the tag, e.g. 'f' is returned
        for tag 'f:foobar'.
        """
        return self._pfx

    @property
    def nsuri(self):
        """
        Return the namespace URI for the tag, e.g. 'http://foo/bar' would be
        returned for tag 'f:foobar' if the 'f' prefix maps to
        'http://foo/bar' in nsmap.
        """
        return self._ns_uri


def nsdecls(*prefixes):
    """
    Return a string containing a namespace declaration for each of the
    namespace prefix strings, e.g. 'p', 'ct', passed as *prefixes*.
    """
    return ' '.join(['xmlns:%s="%s"' % (pfx, nsmap[pfx]) for pfx in prefixes])


def nspfxmap(*nspfxs):
    """
    Return a dict containing the subset namespace prefix mappings specified by
    *nspfxs*. Any number of namespace prefixes can be supplied, e.g.
    namespaces('a', 'r', 'p').
    """
    return dict((pfx, nsmap[pfx]) for pfx in nspfxs)


def qn(tag):
    """
    Stands for "qualified name", a utility function to turn a namespace
    prefixed tag name into a Clark-notation qualified tag name for lxml. For
    example, ``qn('p:cSld')`` returns ``'{http://schemas.../main}cSld'``.
    """
    prefix, tagroot = tag.split(':')
    uri = nsmap[prefix]
    return '{%s}%s' % (uri, tagroot)



# -*- coding: utf-8 -*-

from . import color
from .command import Command as _Command
from .log import Logger as _Logger


class Logger(_Logger):
    def message(self, level, *args):
        msg = ' '.join((str(o) for o in args))

        if level == 'start':
            return color.bold(msg)
        _ = {
            'debug': 'gray',
            'info': 'green',
            'warn': 'yellow',
            'error': 'red',
            'end': 'white',
        }

        if level == 'error':
            msg = color.red(msg)

        if level in _:
            fn = getattr(color, _[level])
            return '%s %s' % (fn('*'), msg)
        return msg


log = Logger()


class Command(_Command):
    def print_title(self, title):
        if 'Option' in title:
            print(color.green(title))
        elif 'Command' in title:
            print(color.magenta(title))
        else:
            print(title)
        return self

    def add_log_options(self, verbose_func=None, quiet_func=None):
        """
        A helper for setting up log options
        """

        if not verbose_func:
            def verbose_func():
                return log.config(verbose=True)

        if not quiet_func:
            def quiet_func():
                return log.config(quiet=True)

        self.option('-v, --verbose', 'show more logs', verbose_func)
        self.option('-q, --quiet', 'show less logs', quiet_func)
        return self



import time
from netmiko.base_connection import BaseConnection


class PluribusSSH(BaseConnection):
    """Common methods for Pluribus."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config_mode = False

    def disable_paging(self, command="pager off", delay_factor=1):
        """Make sure paging is disabled."""
        return super().disable_paging(command=command, delay_factor=delay_factor)

    def session_preparation(self):
        """Prepare the netmiko session."""
        self._test_channel_read()
        self.set_base_prompt()
        self.disable_paging()
        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def check_config_mode(self, *args, **kwargs):
        """
        Pluribus devices don't have a config mode.
        Therefore it can be considered as always in config mode.
        """
        return self._config_mode

    def config_mode(self, *args, **kwargs):
        """No special actions to enter in config mode."""
        self._config_mode = True
        return ""

    def exit_config_mode(self, *args, **kwargs):
        """No special actions to exit config mode."""
        self._config_mode = False
        return ""



# Copyright (C) 2006-2007  Robey Pointer <robeypointer@gmail.com>
# Copyright (C) 2012  Olle Lundberg <geek@nerd.sh>
#
# This file is part of paramiko.
#
# Paramiko is free software; you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# Paramiko is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Paramiko; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

"""
Configuration file (aka ``ssh_config``) support.
"""

import fnmatch
import getpass
import os
import re
import shlex
import socket
from functools import partial

from .py3compat import StringIO

invoke, invoke_import_error = None, None
try:
    import invoke
except ImportError as e:
    invoke_import_error = e

from .ssh_exception import CouldNotCanonicalize, ConfigParseError


SSH_PORT = 22


class SSHConfig(object):
    """
    Representation of config information as stored in the format used by
    OpenSSH. Queries can be made via `lookup`. The format is described in
    OpenSSH's ``ssh_config`` man page. This class is provided primarily as a
    convenience to posix users (since the OpenSSH format is a de-facto
    standard on posix) but should work fine on Windows too.

    .. versionadded:: 1.6
    """

    SETTINGS_REGEX = re.compile(r"(\w+)(?:\s*=\s*|\s+)(.+)")

    # TODO: do a full scan of ssh.c & friends to make sure we're fully
    # compatible across the board, e.g. OpenSSH 8.1 added %n to ProxyCommand.
    TOKENS_BY_CONFIG_KEY = {
        "controlpath": ["%h", "%l", "%L", "%n", "%p", "%r", "%u"],
        "hostname": ["%h"],
        "identityfile": ["~", "%d", "%h", "%l", "%u", "%r"],
        "proxycommand": ["~", "%h", "%p", "%r"],
        # Doesn't seem worth making this 'special' for now, it will fit well
        # enough (no actual match-exec config key to be confused with).
        "match-exec": ["%d", "%h", "%L", "%l", "%n", "%p", "%r", "%u"],
    }

    def __init__(self):
        """
        Create a new OpenSSH config object.

        Note: the newer alternate constructors `from_path`, `from_file` and
        `from_text` are simpler to use, as they parse on instantiation. For
        example, instead of::

            config = SSHConfig()
            config.parse(open("some-path.config")

        you could::

            config = SSHConfig.from_file(open("some-path.config"))
            # Or more directly:
            config = SSHConfig.from_path("some-path.config")
            # Or if you have arbitrary ssh_config text from some other source:
            config = SSHConfig.from_text("Host foo\\n\\tUser bar")
        """
        self._config = []

    @classmethod
    def from_text(cls, text):
        """
        Create a new, parsed `SSHConfig` from ``text`` string.

        .. versionadded:: 2.7
        """
        return cls.from_file(StringIO(text))

    @classmethod
    def from_path(cls, path):
        """
        Create a new, parsed `SSHConfig` from the file found at ``path``.

        .. versionadded:: 2.7
        """
        with open(path) as flo:
            return cls.from_file(flo)

    @classmethod
    def from_file(cls, flo):
        """
        Create a new, parsed `SSHConfig` from file-like object ``flo``.

        .. versionadded:: 2.7
        """
        obj = cls()
        obj.parse(flo)
        return obj

    def parse(self, file_obj):
        """
        Read an OpenSSH config from the given file object.

        :param file_obj: a file-like object to read the config file from
        """
        # Start out w/ implicit/anonymous global host-like block to hold
        # anything not contained by an explicit one.
        context = {"host": ["*"], "config": {}}
        for line in file_obj:
            # Strip any leading or trailing whitespace from the line.
            # Refer to https://github.com/paramiko/paramiko/issues/499
            line = line.strip()
            # Skip blanks, comments
            if not line or line.startswith("#"):
                continue

            # Parse line into key, value
            match = re.match(self.SETTINGS_REGEX, line)
            if not match:
                raise ConfigParseError("Unparsable line {}".format(line))
            key = match.group(1).lower()
            value = match.group(2)

            # Host keyword triggers switch to new block/context
            if key in ("host", "match"):
                self._config.append(context)
                context = {"config": {}}
                if key == "host":
                    # TODO 3.0: make these real objects or at least name this
                    # "hosts" to acknowledge it's an iterable. (Doing so prior
                    # to 3.0, despite it being a private API, feels bad -
                    # surely such an old codebase has folks actually relying on
                    # these keys.)
                    context["host"] = self._get_hosts(value)
                else:
                    context["matches"] = self._get_matches(value)
            # Special-case for noop ProxyCommands
            elif key == "proxycommand" and value.lower() == "none":
                # Store 'none' as None; prior to 3.x, it will get stripped out
                # at the end (for compatibility with issue #415). After 3.x, it
                # will simply not get stripped, leaving a nice explicit marker.
                context["config"][key] = None
            # All other keywords get stored, directly or via append
            else:
                if value.startswith('"') and value.endswith('"'):
                    value = value[1:-1]

                # identityfile, localforward, remoteforward keys are special
                # cases, since they are allowed to be specified multiple times
                # and they should be tried in order of specification.
                if key in ["identityfile", "localforward", "remoteforward"]:
                    if key in context["config"]:
                        context["config"][key].append(value)
                    else:
                        context["config"][key] = [value]
                elif key not in context["config"]:
                    context["config"][key] = value
        # Store last 'open' block and we're done
        self._config.append(context)

    def lookup(self, hostname):
        """
        Return a dict (`SSHConfigDict`) of config options for a given hostname.

        The host-matching rules of OpenSSH's ``ssh_config`` man page are used:
        For each parameter, the first obtained value will be used.  The
        configuration files contain sections separated by ``Host`` and/or
        ``Match`` specifications, and that section is only applied for hosts
        which match the given patterns or keywords

        Since the first obtained value for each parameter is used, more host-
        specific declarations should be given near the beginning of the file,
        and general defaults at the end.

        The keys in the returned dict are all normalized to lowercase (look for
        ``"port"``, not ``"Port"``. The values are processed according to the
        rules for substitution variable expansion in ``ssh_config``.

        Finally, please see the docs for `SSHConfigDict` for deeper info on
        features such as optional type conversion methods, e.g.::

            conf = my_config.lookup('myhost')
            assert conf['passwordauthentication'] == 'yes'
            assert conf.as_bool('passwordauthentication') is True

        .. note::
            If there is no explicitly configured ``HostName`` value, it will be
            set to the being-looked-up hostname, which is as close as we can
            get to OpenSSH's behavior around that particular option.

        :param str hostname: the hostname to lookup

        .. versionchanged:: 2.5
            Returns `SSHConfigDict` objects instead of dict literals.
        .. versionchanged:: 2.7
            Added canonicalization support.
        .. versionchanged:: 2.7
            Added ``Match`` support.
        """
        # First pass
        options = self._lookup(hostname=hostname)
        # Inject HostName if it was not set (this used to be done incidentally
        # during tokenization, for some reason).
        if "hostname" not in options:
            options["hostname"] = hostname
        # Handle canonicalization
        canon = options.get("canonicalizehostname", None) in ("yes", "always")
        maxdots = int(options.get("canonicalizemaxdots", 1))
        if canon and hostname.count(".") <= maxdots:
            # NOTE: OpenSSH manpage does not explicitly state this, but its
            # implementation for CanonicalDomains is 'split on any whitespace'.
            domains = options["canonicaldomains"].split()
            hostname = self.canonicalize(hostname, options, domains)
            # Overwrite HostName again here (this is also what OpenSSH does)
            options["hostname"] = hostname
            options = self._lookup(hostname, options, canonical=True)
        return options

    def _lookup(self, hostname, options=None, canonical=False):
        # Init
        if options is None:
            options = SSHConfigDict()
        # Iterate all stanzas, applying any that match, in turn (so that things
        # like Match can reference currently understood state)
        for context in self._config:
            if not (
                self._pattern_matches(context.get("host", []), hostname)
                or self._does_match(
                    context.get("matches", []), hostname, canonical, options
                )
            ):
                continue
            for key, value in context["config"].items():
                if key not in options:
                    # Create a copy of the original value,
                    # else it will reference the original list
                    # in self._config and update that value too
                    # when the extend() is being called.
                    options[key] = value[:] if value is not None else value
                elif key == "identityfile":
                    options[key].extend(
                        x for x in value if x not in options[key]
                    )
        # Expand variables in resulting values (besides 'Match exec' which was
        # already handled above)
        options = self._expand_variables(options, hostname)
        # TODO: remove in 3.x re #670
        if "proxycommand" in options and options["proxycommand"] is None:
            del options["proxycommand"]
        return options

    def canonicalize(self, hostname, options, domains):
        """
        Return canonicalized version of ``hostname``.

        :param str hostname: Target hostname.
        :param options: An `SSHConfigDict` from a previous lookup pass.
        :param domains: List of domains (e.g. ``["paramiko.org"]``).

        :returns: A canonicalized hostname if one was found, else ``None``.

        .. versionadded:: 2.7
        """
        found = False
        for domain in domains:
            candidate = "{}.{}".format(hostname, domain)
            family_specific = _addressfamily_host_lookup(candidate, options)
            if family_specific is not None:
                # TODO: would we want to dig deeper into other results? e.g. to
                # find something that satisfies PermittedCNAMEs when that is
                # implemented?
                found = family_specific[0]
            else:
                # TODO: what does ssh use here and is there a reason to use
                # that instead of gethostbyname?
                try:
                    found = socket.gethostbyname(candidate)
                except socket.gaierror:
                    pass
            if found:
                # TODO: follow CNAME (implied by found != candidate?) if
                # CanonicalizePermittedCNAMEs allows it
                return candidate
        # If we got here, it means canonicalization failed.
        # When CanonicalizeFallbackLocal is undefined or 'yes', we just spit
        # back the original hostname.
        if options.get("canonicalizefallbacklocal", "yes") == "yes":
            return hostname
        # And here, we failed AND fallback was set to a non-yes value, so we
        # need to get mad.
        raise CouldNotCanonicalize(hostname)

    def get_hostnames(self):
        """
        Return the set of literal hostnames defined in the SSH config (both
        explicit hostnames and wildcard entries).
        """
        hosts = set()
        for entry in self._config:
            hosts.update(entry["host"])
        return hosts

    def _pattern_matches(self, patterns, target):
        # Convenience auto-splitter if not already a list
        if hasattr(patterns, "split"):
            patterns = patterns.split(",")
        match = False
        for pattern in patterns:
            # Short-circuit if target matches a negated pattern
            if pattern.startswith("!") and fnmatch.fnmatch(
                target, pattern[1:]
            ):
                return False
            # Flag a match, but continue (in case of later negation) if regular
            # match occurs
            elif fnmatch.fnmatch(target, pattern):
                match = True
        return match

    # TODO 3.0: remove entirely (is now unused internally)
    def _allowed(self, hosts, hostname):
        return self._pattern_matches(hosts, hostname)

    def _does_match(self, match_list, target_hostname, canonical, options):
        matched = []
        candidates = match_list[:]
        local_username = getpass.getuser()
        while candidates:
            candidate = candidates.pop(0)
            passed = None
            # Obtain latest host/user value every loop, so later Match may
            # reference values assigned within a prior Match.
            configured_host = options.get("hostname", None)
            configured_user = options.get("user", None)
            type_, param = candidate["type"], candidate["param"]
            # Canonical is a hard pass/fail based on whether this is a
            # canonicalized re-lookup.
            if type_ == "canonical":
                if self._should_fail(canonical, candidate):
                    return False
            # The parse step ensures we only see this by itself or after
            # canonical, so it's also an easy hard pass. (No negation here as
            # that would be uh, pretty weird?)
            elif type_ == "all":
                return True
            # From here, we are testing various non-hard criteria,
            # short-circuiting only on fail
            elif type_ == "host":
                hostval = configured_host or target_hostname
                passed = self._pattern_matches(param, hostval)
            elif type_ == "originalhost":
                passed = self._pattern_matches(param, target_hostname)
            elif type_ == "user":
                user = configured_user or local_username
                passed = self._pattern_matches(param, user)
            elif type_ == "localuser":
                passed = self._pattern_matches(param, local_username)
            elif type_ == "exec":
                exec_cmd = self._tokenize(
                    options, target_hostname, "match-exec", param
                )
                # This is the laziest spot in which we can get mad about an
                # inability to import Invoke.
                if invoke is None:
                    raise invoke_import_error
                # Like OpenSSH, we 'redirect' stdout but let stderr bubble up
                passed = invoke.run(exec_cmd, hide="stdout", warn=True).ok
            # Tackle any 'passed, but was negated' results from above
            if passed is not None and self._should_fail(passed, candidate):
                return False
            # Made it all the way here? Everything matched!
            matched.append(candidate)
        # Did anything match? (To be treated as bool, usually.)
        return matched

    def _should_fail(self, would_pass, candidate):
        return would_pass if candidate["negate"] else not would_pass

    def _tokenize(self, config, target_hostname, key, value):
        """
        Tokenize a string based on current config/hostname data.

        :param config: Current config data.
        :param target_hostname: Original target connection hostname.
        :param key: Config key being tokenized (used to filter token list).
        :param value: Config value being tokenized.

        :returns: The tokenized version of the input ``value`` string.
        """
        allowed_tokens = self._allowed_tokens(key)
        # Short-circuit if no tokenization possible
        if not allowed_tokens:
            return value
        # Obtain potentially configured hostname, for use with %h.
        # Special-case where we are tokenizing the hostname itself, to avoid
        # replacing %h with a %h-bearing value, etc.
        configured_hostname = target_hostname
        if key != "hostname":
            configured_hostname = config.get("hostname", configured_hostname)
        # Ditto the rest of the source values
        if "port" in config:
            port = config["port"]
        else:
            port = SSH_PORT
        user = getpass.getuser()
        if "user" in config:
            remoteuser = config["user"]
        else:
            remoteuser = user
        local_hostname = socket.gethostname().split(".")[0]
        local_fqdn = LazyFqdn(config, local_hostname)
        homedir = os.path.expanduser("~")
        # The actual tokens!
        replacements = {
            # TODO: %%???
            # TODO: %C?
            "%d": homedir,
            "%h": configured_hostname,
            # TODO: %i?
            "%L": local_hostname,
            "%l": local_fqdn,
            # also this is pseudo buggy when not in Match exec mode so document
            # that. also WHY is that the case?? don't we do all of this late?
            "%n": target_hostname,
            "%p": port,
            "%r": remoteuser,
            # TODO: %T? don't believe this is possible however
            "%u": user,
            "~": homedir,
        }
        # Do the thing with the stuff
        tokenized = value
        for find, replace in replacements.items():
            if find not in allowed_tokens:
                continue
            tokenized = tokenized.replace(find, str(replace))
        # TODO: log? eg that value -> tokenized
        return tokenized

    def _allowed_tokens(self, key):
        """
        Given config ``key``, return list of token strings to tokenize.

        .. note::
            This feels like it wants to eventually go away, but is used to
            preserve as-strict-as-possible compatibility with OpenSSH, which
            for whatever reason only applies some tokens to some config keys.
        """
        return self.TOKENS_BY_CONFIG_KEY.get(key, [])

    def _expand_variables(self, config, target_hostname):
        """
        Return a dict of config options with expanded substitutions
        for a given original & current target hostname.

        Please refer to :doc:`/api/config` for details.

        :param dict config: the currently parsed config
        :param str hostname: the hostname whose config is being looked up
        """
        for k in config:
            if config[k] is None:
                continue
            tokenizer = partial(self._tokenize, config, target_hostname, k)
            if isinstance(config[k], list):
                for i, value in enumerate(config[k]):
                    config[k][i] = tokenizer(value)
            else:
                config[k] = tokenizer(config[k])
        return config

    def _get_hosts(self, host):
        """
        Return a list of host_names from host value.
        """
        try:
            return shlex.split(host)
        except ValueError:
            raise ConfigParseError("Unparsable host {}".format(host))

    def _get_matches(self, match):
        """
        Parse a specific Match config line into a list-of-dicts for its values.

        Performs some parse-time validation as well.
        """
        matches = []
        tokens = shlex.split(match)
        while tokens:
            match = {"type": None, "param": None, "negate": False}
            type_ = tokens.pop(0)
            # Handle per-keyword negation
            if type_.startswith("!"):
                match["negate"] = True
                type_ = type_[1:]
            match["type"] = type_
            # all/canonical have no params (everything else does)
            if type_ in ("all", "canonical"):
                matches.append(match)
                continue
            if not tokens:
                raise ConfigParseError(
                    "Missing parameter to Match '{}' keyword".format(type_)
                )
            match["param"] = tokens.pop(0)
            matches.append(match)
        # Perform some (easier to do now than in the middle) validation that is
        # better handled here than at lookup time.
        keywords = [x["type"] for x in matches]
        if "all" in keywords:
            allowable = ("all", "canonical")
            ok, bad = (
                list(filter(lambda x: x in allowable, keywords)),
                list(filter(lambda x: x not in allowable, keywords)),
            )
            err = None
            if any(bad):
                err = "Match does not allow 'all' mixed with anything but 'canonical'"  # noqa
            elif "canonical" in ok and ok.index("canonical") > ok.index("all"):
                err = "Match does not allow 'all' before 'canonical'"
            if err is not None:
                raise ConfigParseError(err)
        return matches


def _addressfamily_host_lookup(hostname, options):
    """
    Try looking up ``hostname`` in an IPv4 or IPv6 specific manner.

    This is an odd duck due to needing use in two divergent use cases. It looks
    up ``AddressFamily`` in ``options`` and if it is ``inet`` or ``inet6``,
    this function uses `socket.getaddrinfo` to perform a family-specific
    lookup, returning the result if successful.

    In any other situation -- lookup failure, or ``AddressFamily`` being
    unspecified or ``any`` -- ``None`` is returned instead and the caller is
    expected to do something situation-appropriate like calling
    `socket.gethostbyname`.

    :param str hostname: Hostname to look up.
    :param options: `SSHConfigDict` instance w/ parsed options.
    :returns: ``getaddrinfo``-style tuples, or ``None``, depending.
    """
    address_family = options.get("addressfamily", "any").lower()
    if address_family == "any":
        return
    try:
        family = socket.AF_INET6
        if address_family == "inet":
            family = socket.AF_INET
        return socket.getaddrinfo(
            hostname,
            None,
            family,
            socket.SOCK_DGRAM,
            socket.IPPROTO_IP,
            socket.AI_CANONNAME,
        )
    except socket.gaierror:
        pass


class LazyFqdn(object):
    """
    Returns the host's fqdn on request as string.
    """

    def __init__(self, config, host=None):
        self.fqdn = None
        self.config = config
        self.host = host

    def __str__(self):
        if self.fqdn is None:
            #
            # If the SSH config contains AddressFamily, use that when
            # determining  the local host's FQDN. Using socket.getfqdn() from
            # the standard library is the most general solution, but can
            # result in noticeable delays on some platforms when IPv6 is
            # misconfigured or not available, as it calls getaddrinfo with no
            # address family specified, so both IPv4 and IPv6 are checked.
            #

            # Handle specific option
            fqdn = None
            results = _addressfamily_host_lookup(self.host, self.config)
            if results is not None:
                for res in results:
                    af, socktype, proto, canonname, sa = res
                    if canonname and "." in canonname:
                        fqdn = canonname
                        break
            # Handle 'any' / unspecified / lookup failure
            if fqdn is None:
                fqdn = socket.getfqdn()
            # Cache
            self.fqdn = fqdn
        return self.fqdn


class SSHConfigDict(dict):
    """
    A dictionary wrapper/subclass for per-host configuration structures.

    This class introduces some usage niceties for consumers of `SSHConfig`,
    specifically around the issue of variable type conversions: normal value
    access yields strings, but there are now methods such as `as_bool` and
    `as_int` that yield casted values instead.

    For example, given the following ``ssh_config`` file snippet::

        Host foo.example.com
            PasswordAuthentication no
            Compression yes
            ServerAliveInterval 60

    the following code highlights how you can access the raw strings as well as
    usefully Python type-casted versions (recalling that keys are all
    normalized to lowercase first)::

        my_config = SSHConfig()
        my_config.parse(open('~/.ssh/config'))
        conf = my_config.lookup('foo.example.com')

        assert conf['passwordauthentication'] == 'no'
        assert conf.as_bool('passwordauthentication') is False
        assert conf['compression'] == 'yes'
        assert conf.as_bool('compression') is True
        assert conf['serveraliveinterval'] == '60'
        assert conf.as_int('serveraliveinterval') == 60

    .. versionadded:: 2.5
    """

    def __init__(self, *args, **kwargs):
        # Hey, guess what? Python 2's userdict is an old-style class!
        super(SSHConfigDict, self).__init__(*args, **kwargs)

    def as_bool(self, key):
        """
        Express given key's value as a boolean type.

        Typically, this is used for ``ssh_config``'s pseudo-boolean values
        which are either ``"yes"`` or ``"no"``. In such cases, ``"yes"`` yields
        ``True`` and any other value becomes ``False``.

        .. note::
            If (for whatever reason) the stored value is already boolean in
            nature, it's simply returned.

        .. versionadded:: 2.5
        """
        val = self[key]
        if isinstance(val, bool):
            return val
        return val.lower() == "yes"

    def as_int(self, key):
        """
        Express given key's value as an integer, if possible.

        This method will raise ``ValueError`` or similar if the value is not
        int-appropriate, same as the builtin `int` type.

        .. versionadded:: 2.5
        """
        return int(self[key])



# encoding: utf-8

"""
Tabstop-related proxy types.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from ..shared import ElementProxy
from docx.enum.text import WD_TAB_ALIGNMENT, WD_TAB_LEADER


class TabStops(ElementProxy):
    """
    A sequence of |TabStop| objects providing access to the tab stops of
    a paragraph or paragraph style. Supports iteration, indexed access, del,
    and len(). It is accesed using the :attr:`~.ParagraphFormat.tab_stops`
    property of ParagraphFormat; it is not intended to be constructed
    directly.
    """

    __slots__ = ('_pPr')

    def __init__(self, element):
        super(TabStops, self).__init__(element, None)
        self._pPr = element

    def __delitem__(self, idx):
        """
        Remove the tab at offset *idx* in this sequence.
        """
        tabs = self._pPr.tabs
        try:
            tabs.remove(tabs[idx])
        except (AttributeError, IndexError):
            raise IndexError('tab index out of range')

        if len(tabs) == 0:
            self._pPr.remove(tabs)

    def __getitem__(self, idx):
        """
        Enables list-style access by index.
        """
        tabs = self._pPr.tabs
        if tabs is None:
            raise IndexError('TabStops object is empty')
        tab = tabs.tab_lst[idx]
        return TabStop(tab)

    def __iter__(self):
        """
        Generate a TabStop object for each of the w:tab elements, in XML
        document order.
        """
        tabs = self._pPr.tabs
        if tabs is not None:
            for tab in tabs.tab_lst:
                yield TabStop(tab)

    def __len__(self):
        tabs = self._pPr.tabs
        if tabs is None:
            return 0
        return len(tabs.tab_lst)

    def add_tab_stop(self, position, alignment=WD_TAB_ALIGNMENT.LEFT,
                     leader=WD_TAB_LEADER.SPACES):
        """
        Add a new tab stop at *position*, a |Length| object specifying the
        location of the tab stop relative to the paragraph edge. A negative
        *position* value is valid and appears in hanging indentation. Tab
        alignment defaults to left, but may be specified by passing a member
        of the :ref:`WdTabAlignment` enumeration as *alignment*. An optional
        leader character can be specified by passing a member of the
        :ref:`WdTabLeader` enumeration as *leader*.
        """
        tabs = self._pPr.get_or_add_tabs()
        tab = tabs.insert_tab_in_order(position, alignment, leader)
        return TabStop(tab)

    def clear_all(self):
        """
        Remove all custom tab stops.
        """
        self._pPr._remove_tabs()


class TabStop(ElementProxy):
    """
    An individual tab stop applying to a paragraph or style. Accessed using
    list semantics on its containing |TabStops| object.
    """

    __slots__ = ('_tab')

    def __init__(self, element):
        super(TabStop, self).__init__(element, None)
        self._tab = element

    @property
    def alignment(self):
        """
        A member of :ref:`WdTabAlignment` specifying the alignment setting
        for this tab stop. Read/write.
        """
        return self._tab.val

    @alignment.setter
    def alignment(self, value):
        self._tab.val = value

    @property
    def leader(self):
        """
        A member of :ref:`WdTabLeader` specifying a repeating character used
        as a "leader", filling in the space spanned by this tab. Assigning
        |None| produces the same result as assigning `WD_TAB_LEADER.SPACES`.
        Read/write.
        """
        return self._tab.leader

    @leader.setter
    def leader(self, value):
        self._tab.leader = value

    @property
    def position(self):
        """
        A |Length| object representing the distance of this tab stop from the
        inside edge of the paragraph. May be positive or negative.
        Read/write.
        """
        return self._tab.pos

    @position.setter
    def position(self, value):
        tab = self._tab
        tabs = tab.getparent()
        self._tab = tabs.insert_tab_in_order(value, tab.val, tab.leader)
        tabs.remove(tab)



#!/usr/bin/env python
# flake8: noqa

"""
    terminal
    ~~~~~~~~

    A terminal environment tools.

    :copyright: (c) 2013 by Hsiaoming Yang.
"""

from __future__ import absolute_import

from .color import *
from .prompt import *
from .log import Logger
from .command import Command, Option

log = Logger()

__homepage__ = 'https://github.com/lepture/terminal'
__author__ = 'Hsiaoming Yang <me@lepture.com>'
__version__ = '0.4.0'



"""
Click is a simple Python module inspired by the stdlib optparse to make
writing command line scripts fun. Unlike other modules, it's based
around a simple API that does not come with too much magic and is
composable.
"""
from .core import Argument
from .core import BaseCommand
from .core import Command
from .core import CommandCollection
from .core import Context
from .core import Group
from .core import MultiCommand
from .core import Option
from .core import Parameter
from .decorators import argument
from .decorators import command
from .decorators import confirmation_option
from .decorators import group
from .decorators import help_option
from .decorators import make_pass_decorator
from .decorators import option
from .decorators import pass_context
from .decorators import pass_obj
from .decorators import password_option
from .decorators import version_option
from .exceptions import Abort
from .exceptions import BadArgumentUsage
from .exceptions import BadOptionUsage
from .exceptions import BadParameter
from .exceptions import ClickException
from .exceptions import FileError
from .exceptions import MissingParameter
from .exceptions import NoSuchOption
from .exceptions import UsageError
from .formatting import HelpFormatter
from .formatting import wrap_text
from .globals import get_current_context
from .parser import OptionParser
from .termui import clear
from .termui import confirm
from .termui import echo_via_pager
from .termui import edit
from .termui import get_terminal_size
from .termui import getchar
from .termui import launch
from .termui import pause
from .termui import progressbar
from .termui import prompt
from .termui import secho
from .termui import style
from .termui import unstyle
from .types import BOOL
from .types import Choice
from .types import DateTime
from .types import File
from .types import FLOAT
from .types import FloatRange
from .types import INT
from .types import IntRange
from .types import ParamType
from .types import Path
from .types import STRING
from .types import Tuple
from .types import UNPROCESSED
from .types import UUID
from .utils import echo
from .utils import format_filename
from .utils import get_app_dir
from .utils import get_binary_stream
from .utils import get_os_args
from .utils import get_text_stream
from .utils import open_file

# Controls if click should emit the warning about the use of unicode
# literals.
disable_unicode_literals_warning = False

__version__ = "7.1.1"



###############################################################################
#
# Custom - A class for writing the Excel XLSX Custom Property file.
#
# Copyright 2013-2020, John McNamara, jmcnamara@cpan.org
#

# Package imports.
from . import xmlwriter


class Custom(xmlwriter.XMLwriter):
    """
    A class for writing the Excel XLSX Custom Workbook Property file.


    """

    ###########################################################################
    #
    # Public API.
    #
    ###########################################################################

    def __init__(self):
        """
        Constructor.

        """

        super(Custom, self).__init__()

        self.properties = []
        self.pid = 1

    def _set_properties(self, properties):
        # Set the document properties.
        self.properties = properties

    ###########################################################################
    #
    # Private API.
    #
    ###########################################################################

    def _assemble_xml_file(self):
        # Assemble and write the XML file.

        # Write the XML declaration.
        self._xml_declaration()

        self._write_properties()

        self._xml_end_tag('Properties')

        # Close the file.
        self._xml_close()

    ###########################################################################
    #
    # XML methods.
    #
    ###########################################################################

    def _write_properties(self):
        # Write the <Properties> element.
        schema = 'http://schemas.openxmlformats.org/officeDocument/2006/'
        xmlns = schema + 'custom-properties'
        xmlns_vt = schema + 'docPropsVTypes'

        attributes = [
            ('xmlns', xmlns),
            ('xmlns:vt', xmlns_vt),
        ]

        self._xml_start_tag('Properties', attributes)

        for custom_property in self.properties:
            # Write the property element.
            self._write_property(custom_property)

    def _write_property(self, custom_property):
        # Write the <property> element.

        fmtid = '{D5CDD505-2E9C-101B-9397-08002B2CF9AE}'

        name, value, property_type = custom_property
        self.pid += 1

        attributes = [
            ('fmtid', fmtid),
            ('pid', self.pid),
            ('name', name),
        ]

        self._xml_start_tag('property', attributes)

        if property_type == 'number_int':
            # Write the vt:i4 element.
            self._write_vt_i4(value)
        elif property_type == 'number':
            # Write the vt:r8 element.
            self._write_vt_r8(value)
        elif property_type == 'date':
            # Write the vt:filetime element.
            self._write_vt_filetime(value)
        elif property_type == 'bool':
            # Write the vt:bool element.
            self._write_vt_bool(value)
        else:
            # Write the vt:lpwstr element.
            self._write_vt_lpwstr(value)

        self._xml_end_tag('property')

    def _write_vt_lpwstr(self, value):
        # Write the <vt:lpwstr> element.
        self._xml_data_element('vt:lpwstr', value)

    def _write_vt_filetime(self, value):
        # Write the <vt:filetime> element.
        self._xml_data_element('vt:filetime', value)

    def _write_vt_i4(self, value):
        # Write the <vt:i4> element.
        self._xml_data_element('vt:i4', value)

    def _write_vt_r8(self, value):
        # Write the <vt:r8> element.
        self._xml_data_element('vt:r8', value)

    def _write_vt_bool(self, value):
        # Write the <vt:bool> element.

        if value:
            value = 'true'
        else:
            value = 'false'

        self._xml_data_element('vt:bool', value)



"""Extreme support."""
import time
import re
from netmiko.cisco_base_connection import CiscoSSHConnection


class ExtremeExosBase(CiscoSSHConnection):
    """Extreme Exos support.

    Designed for EXOS >= 15.0
    """

    def session_preparation(self):
        self._test_channel_read()
        self.set_base_prompt()
        self.disable_paging(command="disable clipaging")
        self.send_command_timing("disable cli prompting")
        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def set_base_prompt(self, *args, **kwargs):
        """
        Extreme attaches an id to the prompt. The id increases with every command.
        It needs to br stripped off to match the prompt. Eg.

            testhost.1 #
            testhost.2 #
            testhost.3 #

        If new config is loaded and not saved yet, a '* ' prefix appears before the
        prompt, eg.

            * testhost.4 #
            * testhost.5 #
        """
        cur_base_prompt = super().set_base_prompt(*args, **kwargs)
        # Strip off any leading * or whitespace chars; strip off trailing period and digits
        match = re.search(r"[\*\s]*(.*)\.\d+", cur_base_prompt)
        if match:
            self.base_prompt = match.group(1)
            return self.base_prompt
        else:
            return self.base_prompt

    def send_command(self, *args, **kwargs):
        """Extreme needs special handler here due to the prompt changes."""

        # Change send_command behavior to use self.base_prompt
        kwargs.setdefault("auto_find_prompt", False)

        # refresh self.base_prompt
        self.set_base_prompt()
        return super().send_command(*args, **kwargs)

    def config_mode(self, config_command=""):
        """No configuration mode on Extreme Exos."""
        return ""

    def check_config_mode(self, check_string="#"):
        """Checks whether in configuration mode. Returns a boolean."""
        return super().check_config_mode(check_string=check_string)

    def exit_config_mode(self, exit_config=""):
        """No configuration mode on Extreme Exos."""
        return ""

    def save_config(
        self, cmd="save configuration primary", confirm=False, confirm_response=""
    ):
        """Saves configuration."""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )


class ExtremeExosSSH(ExtremeExosBase):
    pass


class ExtremeExosTelnet(ExtremeExosBase):
    def __init__(self, *args, **kwargs):
        default_enter = kwargs.get("default_enter")
        kwargs["default_enter"] = "\r\n" if default_enter is None else default_enter
        super().__init__(*args, **kwargs)



from netmiko.juniper.juniper import JuniperSSH, JuniperTelnet, JuniperFileTransfer
from netmiko.juniper.juniper_screenos import JuniperScreenOsSSH

__all__ = ["JuniperSSH", "JuniperTelnet", "JuniperFileTransfer", "JuniperScreenOsSSH"]



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

import hmac
import warnings

from cryptography import utils
from cryptography.hazmat.bindings._constant_time import lib


if hasattr(hmac, "compare_digest"):
    def bytes_eq(a, b):
        if not isinstance(a, bytes) or not isinstance(b, bytes):
            raise TypeError("a and b must be bytes.")

        return hmac.compare_digest(a, b)

else:
    warnings.warn(
        "Support for your Python version is deprecated. The next version of "
        "cryptography will remove support. Please upgrade to a release "
        "(2.7.7+) that supports hmac.compare_digest as soon as possible.",
        utils.PersistentlyDeprecated2018,
    )

    def bytes_eq(a, b):
        if not isinstance(a, bytes) or not isinstance(b, bytes):
            raise TypeError("a and b must be bytes.")

        return lib.Cryptography_constant_time_bytes_eq(
            a, len(a), b, len(b)
        ) == 1



# encoding: utf-8

"""
Provides mappings that embody aspects of the Open XML spec ISO/IEC 29500.
"""

from .constants import CONTENT_TYPE as CT


default_content_types = (
    ('bin',     CT.PML_PRINTER_SETTINGS),
    ('bin',     CT.SML_PRINTER_SETTINGS),
    ('bin',     CT.WML_PRINTER_SETTINGS),
    ('bmp',     CT.BMP),
    ('emf',     CT.X_EMF),
    ('fntdata', CT.X_FONTDATA),
    ('gif',     CT.GIF),
    ('jpe',     CT.JPEG),
    ('jpeg',    CT.JPEG),
    ('jpg',     CT.JPEG),
    ('png',     CT.PNG),
    ('rels',    CT.OPC_RELATIONSHIPS),
    ('tif',     CT.TIFF),
    ('tiff',    CT.TIFF),
    ('wdp',     CT.MS_PHOTO),
    ('wmf',     CT.X_WMF),
    ('xlsx',    CT.SML_SHEET),
    ('xml',     CT.XML),
)



# encoding: utf-8

"""
Style object hierarchy.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from . import BabelFish
from ..enum.style import WD_STYLE_TYPE
from ..shared import ElementProxy
from ..text.font import Font
from ..text.parfmt import ParagraphFormat


def StyleFactory(style_elm):
    """
    Return a style object of the appropriate |BaseStyle| subclass, according
    to the type of *style_elm*.
    """
    style_cls = {
        WD_STYLE_TYPE.PARAGRAPH: _ParagraphStyle,
        WD_STYLE_TYPE.CHARACTER: _CharacterStyle,
        WD_STYLE_TYPE.TABLE:     _TableStyle,
        WD_STYLE_TYPE.LIST:      _NumberingStyle
    }[style_elm.type]

    return style_cls(style_elm)


class BaseStyle(ElementProxy):
    """
    Base class for the various types of style object, paragraph, character,
    table, and numbering. These properties and methods are inherited by all
    style objects.
    """

    __slots__ = ()

    @property
    def builtin(self):
        """
        Read-only. |True| if this style is a built-in style. |False|
        indicates it is a custom (user-defined) style. Note this value is
        based on the presence of a `customStyle` attribute in the XML, not on
        specific knowledge of which styles are built into Word.
        """
        return not self._element.customStyle

    def delete(self):
        """
        Remove this style definition from the document. Note that calling
        this method does not remove or change the style applied to any
        document content. Content items having the deleted style will be
        rendered using the default style, as is any content with a style not
        defined in the document.
        """
        self._element.delete()
        self._element = None

    @property
    def hidden(self):
        """
        |True| if display of this style in the style gallery and list of
        recommended styles is suppressed. |False| otherwise. In order to be
        shown in the style gallery, this value must be |False| and
        :attr:`.quick_style` must be |True|.
        """
        return self._element.semiHidden_val

    @hidden.setter
    def hidden(self, value):
        self._element.semiHidden_val = value

    @property
    def locked(self):
        """
        Read/write Boolean. |True| if this style is locked. A locked style
        does not appear in the styles panel or the style gallery and cannot
        be applied to document content. This behavior is only active when
        formatting protection is turned on for the document (via the
        Developer menu).
        """
        return self._element.locked_val

    @locked.setter
    def locked(self, value):
        self._element.locked_val = value

    @property
    def name(self):
        """
        The UI name of this style.
        """
        name = self._element.name_val
        if name is None:
            return None
        return BabelFish.internal2ui(name)

    @name.setter
    def name(self, value):
        self._element.name_val = value

    @property
    def priority(self):
        """
        The integer sort key governing display sequence of this style in the
        Word UI. |None| indicates no setting is defined, causing Word to use
        the default value of 0. Style name is used as a secondary sort key to
        resolve ordering of styles having the same priority value.
        """
        return self._element.uiPriority_val

    @priority.setter
    def priority(self, value):
        self._element.uiPriority_val = value

    @property
    def quick_style(self):
        """
        |True| if this style should be displayed in the style gallery when
        :attr:`.hidden` is |False|. Read/write Boolean.
        """
        return self._element.qFormat_val

    @quick_style.setter
    def quick_style(self, value):
        self._element.qFormat_val = value

    @property
    def style_id(self):
        """
        The unique key name (string) for this style. This value is subject to
        rewriting by Word and should generally not be changed unless you are
        familiar with the internals involved.
        """
        return self._element.styleId

    @style_id.setter
    def style_id(self, value):
        self._element.styleId = value

    @property
    def type(self):
        """
        Member of :ref:`WdStyleType` corresponding to the type of this style,
        e.g. ``WD_STYLE_TYPE.PARAGRAPH``.
        """
        type = self._element.type
        if type is None:
            return WD_STYLE_TYPE.PARAGRAPH
        return type

    @property
    def unhide_when_used(self):
        """
        |True| if an application should make this style visible the next time
        it is applied to content. False otherwise. Note that |docx| does not
        automatically unhide a style having |True| for this attribute when it
        is applied to content.
        """
        return self._element.unhideWhenUsed_val

    @unhide_when_used.setter
    def unhide_when_used(self, value):
        self._element.unhideWhenUsed_val = value


class _CharacterStyle(BaseStyle):
    """
    A character style. A character style is applied to a |Run| object and
    primarily provides character-level formatting via the |Font| object in
    its :attr:`.font` property.
    """

    __slots__ = ()

    @property
    def base_style(self):
        """
        Style object this style inherits from or |None| if this style is
        not based on another style.
        """
        base_style = self._element.base_style
        if base_style is None:
            return None
        return StyleFactory(base_style)

    @base_style.setter
    def base_style(self, style):
        style_id = style.style_id if style is not None else None
        self._element.basedOn_val = style_id

    @property
    def font(self):
        """
        The |Font| object providing access to the character formatting
        properties for this style, such as font name and size.
        """
        return Font(self._element)


class _ParagraphStyle(_CharacterStyle):
    """
    A paragraph style. A paragraph style provides both character formatting
    and paragraph formatting such as indentation and line-spacing.
    """

    __slots__ = ()

    def __repr__(self):
        return '_ParagraphStyle(\'%s\') id: %s' % (self.name, id(self))

    @property
    def next_paragraph_style(self):
        """
        |_ParagraphStyle| object representing the style to be applied
        automatically to a new paragraph inserted after a paragraph of this
        style. Returns self if no next paragraph style is defined. Assigning
        |None| or *self* removes the setting such that new paragraphs are
        created using this same style.
        """
        next_style_elm = self._element.next_style
        if next_style_elm is None:
            return self
        if next_style_elm.type != WD_STYLE_TYPE.PARAGRAPH:
            return self
        return StyleFactory(next_style_elm)

    @next_paragraph_style.setter
    def next_paragraph_style(self, style):
        if style is None or style.style_id == self.style_id:
            self._element._remove_next()
        else:
            self._element.get_or_add_next().val = style.style_id

    @property
    def paragraph_format(self):
        """
        The |ParagraphFormat| object providing access to the paragraph
        formatting properties for this style such as indentation.
        """
        return ParagraphFormat(self._element)


class _TableStyle(_ParagraphStyle):
    """
    A table style. A table style provides character and paragraph formatting
    for its contents as well as special table formatting properties.
    """

    __slots__ = ()

    def __repr__(self):
        return '_TableStyle(\'%s\') id: %s' % (self.name, id(self))


class _NumberingStyle(BaseStyle):
    """
    A numbering style. Not yet implemented.
    """

    __slots__ = ()
